'use strict';

var react = require('react');

const defaultReturn = {
    isValid: false,
    value: null,
};
function getRadioValue(options) {
    return Array.isArray(options)
        ? options.reduce((previous, { ref: { checked, value } }) => checked
            ? {
                isValid: true,
                value,
            }
            : previous, defaultReturn)
        : defaultReturn;
}

function getMultipleSelectValue(options) {
    return options.filter(({ selected }) => selected).map(({ value }) => value);
}

function isRadioInput(type) {
    return type === 'radio';
}

function isCheckbox(type) {
    return type === 'checkbox';
}

function getFieldValue(fields, { type, name, options, checked, value }) {
    if (isRadioInput(type)) {
        const result = getRadioValue(fields[name].options).value;
        return result === null ? {} : result;
    }
    if (type === 'select-multiple') {
        return getMultipleSelectValue(options);
    }
    if (isCheckbox(type)) {
        return checked;
    }
    return value;
}

function getFieldsValues(fields, filedName) {
    return Object.values(fields).reduce((previous, data) => {
        const { ref, ref: { name }, } = data;
        const value = getFieldValue(fields, ref);
        if (typeof filedName === 'string') {
            if (name === filedName) {
                return value;
            }
            return previous;
        }
        const copy = Object.assign({}, (previous || {}));
        if (Array.isArray(filedName)) {
            if (filedName.includes(name)) {
                copy[name] = value;
            }
        }
        else {
            copy[name] = value;
        }
        return copy;
    }, {});
}

const DATE_INPUTS = ['date', 'time', 'month', 'datetime', 'datetime-local', 'week'];
const STRING_INPUTS = ['text', 'email', 'password', 'search', 'tel', 'url', 'textarea'];

function getValueAndMessage(item) {
    return {
        value: typeof item === 'object' && item.value ? item.value : item,
        message: typeof item === 'object' && item.message ? item.message : true,
    };
}

var validateField = async ({ ref, ref: { type, value, name, checked }, options, required, maxLength, minLength, min, max, pattern, validate, }, fields) => {
    const copy = {};
    const isRadio = isRadioInput(type);
    if (required &&
        ((isCheckbox(type) && !checked) ||
            (!isRadio && type !== 'checkbox' && value === '') ||
            (isRadio && !getRadioValue(fields[name].options).isValid))) {
        copy[name] = {
            type: 'required',
            message: required,
            // @ts-ignore
            ref: isRadio ? fields[name].options[0].ref : ref,
        };
        return copy;
    }
    if (min || max) {
        let exceedMax;
        let exceedMin;
        const valueNumber = parseFloat(value);
        const { value: maxValue, message: maxMessage } = getValueAndMessage(max);
        const { value: minValue, message: minMessage } = getValueAndMessage(min);
        if (type === 'number') {
            exceedMax = maxValue && valueNumber > maxValue;
            exceedMin = minValue && valueNumber < minValue;
        }
        else if (DATE_INPUTS.includes(type)) {
            if (typeof maxValue === 'string')
                exceedMax = maxValue && new Date(value) > new Date(maxValue);
            if (typeof minValue === 'string')
                exceedMin = minValue && new Date(value) < new Date(minValue);
        }
        if (exceedMax) {
            copy[name] = Object.assign({}, copy[name], { type: 'max', message: maxMessage, ref });
            return copy;
        }
        if (exceedMin) {
            copy[name] = Object.assign({}, copy[name], { type: 'min', message: minMessage, ref });
            return copy;
        }
    }
    if ((maxLength || minLength) && STRING_INPUTS.includes(type)) {
        const { value: maxLengthValue, message: maxLengthMessage } = getValueAndMessage(maxLength);
        const { value: minLengthValue, message: minLengthMessage } = getValueAndMessage(minLength);
        const exceedMax = maxLength && value.toString().length > maxLengthValue;
        const exceedMin = minLength && value.toString().length < minLengthValue;
        if (exceedMax) {
            copy[name] = Object.assign({}, copy[name], { type: 'maxLength', message: maxLengthMessage, ref });
            return copy;
        }
        if (exceedMin) {
            copy[name] = Object.assign({}, copy[name], { type: 'minLength', message: minLengthMessage, ref });
            return copy;
        }
    }
    if (pattern) {
        const { value: patternValue, message: patternMessage } = getValueAndMessage(pattern);
        if (patternValue instanceof RegExp && !patternValue.test(value)) {
            copy[name] = Object.assign({}, copy[name], { type: 'pattern', message: patternMessage, ref });
            return copy;
        }
    }
    if (validate) {
        const fieldValue = isRadio ? getRadioValue(options).value : value;
        if (typeof validate === 'function') {
            const result = await validate(fieldValue);
            if (typeof result !== 'boolean' || !result) {
                copy[name] = Object.assign({}, copy[name], { type: 'validate', message: result || true, ref: isRadio && options ? options[0].ref : ref });
                return copy;
            }
        }
        else if (typeof validate === 'object') {
            const result = await new Promise((resolve) => {
                const values = Object.entries(validate);
                values.reduce(async (previous, [key, validate], index) => {
                    const result = typeof validate === 'function' && (await validate(fieldValue));
                    const lastChild = values.length - 1 === index;
                    if (typeof result !== 'boolean' || !result) {
                        const temp = {
                            type: key,
                            message: result || true,
                        };
                        return lastChild ? resolve(temp) : temp;
                    }
                    return lastChild ? resolve(previous) : previous;
                }, {});
            });
            if (result && Object.keys(result).length) {
                copy[name] = Object.assign({}, copy[name], { ref: isRadio && options ? options[0].ref : ref }, result);
                return copy;
            }
        }
    }
    return copy;
};

var removeAllEventListeners = (ref, validateWithStateUpdate) => {
    ref.removeEventListener('input', validateWithStateUpdate);
    ref.removeEventListener('change', validateWithStateUpdate);
    ref.removeEventListener('blur', validateWithStateUpdate);
};

function findRemovedFieldAndRemoveListener(fields, validateWithStateUpdate, { ref, mutationWatcher, options }, forceDelete = false) {
    if (!ref)
        return;
    const { name, type } = ref;
    if (isRadioInput(type) && options) {
        const optionsCopy = fields[name].options;
        options.forEach(({ ref }, index) => {
            if (!document.body.contains(ref) && fields[name] && optionsCopy && optionsCopy[index]) {
                removeAllEventListeners(optionsCopy[index], validateWithStateUpdate);
                if (optionsCopy[index].mutationWatcher) {
                    // @ts-ignore
                    optionsCopy[index].mutationWatcher.disconnect();
                }
                optionsCopy.splice(index, 1);
            }
        });
        if (Array.isArray(optionsCopy) && !optionsCopy.length) {
            delete fields[name];
            return fields;
        }
    }
    else if (!document.body.contains(ref) || forceDelete) {
        removeAllEventListeners(ref, validateWithStateUpdate);
        if (mutationWatcher)
            mutationWatcher.disconnect();
        delete fields[name];
        return fields;
    }
    return fields;
}

function onDomRemove(element, onDetachCallback) {
    const observer = new MutationObserver(() => {
        function isDetached(el) {
            if (el.parentNode === window.document) {
                return false;
            }
            else if (el.parentNode === null) {
                return true;
            }
            return isDetached(el.parentNode);
        }
        if (isDetached(element)) {
            observer.disconnect();
            onDetachCallback();
        }
    });
    observer.observe(window.document, {
        childList: true,
        subtree: true,
    });
    return observer;
}

function attachEventListeners({ field, validateAndStateUpdate, isRadio, }) {
    const { ref } = field;
    ref.addEventListener(isCheckbox(ref.type) || isRadio ? 'change' : 'input', validateAndStateUpdate);
    ref.addEventListener('blur', validateAndStateUpdate);
}

function parseErrorSchema(error) {
    return error.inner.reduce((previous, current, index) => {
        if (!previous[current.path])
            previous[current.path] = {};
        previous[current.path] = error.errors[index];
        return previous;
    }, {});
}
async function validateWithSchema(ValidationSchema, data) {
    try {
        await ValidationSchema.validate(data, { abortEarly: false });
        return undefined;
    }
    catch (e) {
        return parseErrorSchema(e);
    }
}

function combineFieldValues(data) {
    const output = Object.entries(data).reduce((previous, [key, value]) => {
        const arrayIndex = key.match(/\[\d+\]$/gi);
        if (arrayIndex) {
            const index = arrayIndex[0].slice(1, -1);
            const filedName = key.substr(0, key.indexOf('['));
            if (!previous[filedName])
                previous[filedName] = [];
            previous[filedName][index] = value;
        }
        else {
            previous[key] = value;
        }
        return previous;
    }, {});
    return Object.entries(output).reduce((previous, [key, value]) => {
        if (Array.isArray(value)) {
            previous[key] = value.filter(Boolean);
            return previous;
        }
        previous[key] = value;
        return previous;
    }, {});
}

function shouldUpdateWithError({ errors, name, error, mode, onSubmitModeNotSubmitted, type }) {
    if (onSubmitModeNotSubmitted || (mode === 'onBlur' && type !== 'blur')) {
        return false;
    }
    if (!Object.keys(error).length && !Object.keys(errors).length) {
        return false;
    }
    if (!Object.keys(errors).length && Object.keys(error).length) {
        return true;
    }
    if (!Object.keys(error).length && errors[name]) {
        return true;
    }
    if (Object.keys(error).length) {
        if (!errors[name]) {
            return true;
        }
        else if (errors[name].type !== error[name].type || errors[name].message !== error[name].message) {
            return true;
        }
    }
    return false;
}

var warnMissingRef = (payload) => console.warn(`Oops missing the name for field: ${payload}`);

function useForm({ mode, validationSchema } = {
    mode: 'onSubmit',
}) {
    const fieldsRef = react.useRef({});
    const errorsRef = react.useRef({});
    const isWatchAllRef = react.useRef(false);
    const submitCountRef = react.useRef(0);
    const isSubmittingRef = react.useRef(false);
    const isSubmittedRef = react.useRef(false);
    const isDirtyRef = react.useRef(false);
    const touchedFieldsRef = react.useRef([]);
    const watchFieldsRef = react.useRef({});
    const [errors, reRenderForm] = react.useState({});
    async function validateAndStateUpdate({ target: { name }, type }) {
        const fields = fieldsRef.current;
        const errors = errorsRef.current;
        const ref = fields[name];
        const onSubmitModeNotSubmitted = !isSubmittedRef.current && (mode === 'onSubmit' || !mode);
        const isWatchAll = isWatchAllRef.current;
        const shouldUpdateWatchMode = isWatchAll || watchFieldsRef.current[name];
        const shouldUpdateValidateMode = mode === 'onChange' || (mode === 'onBlur' && type === 'blur');
        let shouldUpdateState = isWatchAll;
        if (!isDirtyRef.current) {
            isDirtyRef.current = true;
            shouldUpdateState = true;
        }
        if (!touchedFieldsRef.current.includes(name)) {
            touchedFieldsRef.current.push(name);
            shouldUpdateState = true;
        }
        if (onSubmitModeNotSubmitted && shouldUpdateWatchMode)
            return reRenderForm({});
        if (validationSchema) {
            const result = getFieldsValues(fields);
            const error = (await validateWithSchema(validationSchema, result)) || {};
            const shouldUpdate = ((!error[name] && errors[name]) || error[name]) && shouldUpdateValidateMode;
            if (shouldUpdate || shouldUpdateWatchMode) {
                const errorsCopy = Object.assign({}, errors, { [name]: error[name] });
                if (!error[name])
                    delete errorsCopy[name];
                errorsRef.current = errorsCopy;
                return reRenderForm(errorsCopy);
            }
        }
        else {
            const error = await validateField(ref, fields);
            const shouldUpdate = shouldUpdateWithError({
                errors,
                error,
                onSubmitModeNotSubmitted,
                name,
                mode,
                type,
            });
            if (shouldUpdate || shouldUpdateValidateMode || shouldUpdateWatchMode) {
                const errorsCopy = Object.assign({}, errors, error);
                if (!error[name])
                    delete errorsCopy[name];
                errorsRef.current = errorsCopy;
                return reRenderForm(errorsCopy);
            }
        }
        if (shouldUpdateState)
            reRenderForm(errors);
    }
    const removeEventListener = findRemovedFieldAndRemoveListener.bind(null, fieldsRef.current, validateAndStateUpdate);
    function registerIntoAllFields(elementRef, data = { required: false, validate: undefined }) {
        if (elementRef && !elementRef.name)
            return warnMissingRef(elementRef);
        const { name, type, value } = elementRef;
        const { required, validate } = data;
        const inputData = Object.assign({}, data, { ref: elementRef });
        const fields = fieldsRef.current;
        const isRadio = isRadioInput(type);
        const field = fields[name];
        const existRadioOptionIndex = isRadio && field && Array.isArray(field.options)
            ? field.options.findIndex(({ ref }) => value === ref.value)
            : -1;
        if ((!isRadio && field) || (isRadio && existRadioOptionIndex > -1))
            return;
        if (isRadio) {
            if (!field)
                fields[name] = { options: [], required, validate, ref: { type: 'radio', name } };
            if (validate)
                fields[name].validate = validate;
            (fields[name].options || []).push(Object.assign({}, inputData, { mutationWatcher: onDomRemove(elementRef, () => removeEventListener(inputData, true)) }));
        }
        else {
            fields[name] = Object.assign({}, inputData, { mutationWatcher: onDomRemove(elementRef, () => removeEventListener(inputData, true)) });
        }
        attachEventListeners({
            field: isRadio ? (fields[name].options || [])[(fields[name].options || []).length - 1] : fields[name],
            isRadio,
            validateAndStateUpdate,
        });
    }
    function watch(filedNames, defaultValue) {
        const watchFields = watchFieldsRef.current;
        if (typeof filedNames === 'string') {
            watchFields[filedNames] = true;
        }
        else if (Array.isArray(filedNames)) {
            filedNames.forEach((name) => {
                watchFields[name] = true;
            });
        }
        else {
            isWatchAllRef.current = true;
            watchFieldsRef.current = {};
        }
        const values = getFieldsValues(fieldsRef.current, filedNames);
        const result = values === undefined || (values.constructor === Object && Object.keys(values).length === 0) ? undefined : values;
        return result === undefined ? defaultValue : result;
    }
    function register(data) {
        if (!data)
            return;
        if (data.type) {
            if (!data.name) {
                warnMissingRef(data);
                return;
            }
            registerIntoAllFields(data);
        }
        return (ref) => ref && registerIntoAllFields(ref, data);
    }
    const handleSubmit = (callback) => async (e) => {
        if (e) {
            e.preventDefault();
            e.persist();
        }
        let fieldErrors;
        let fieldValues;
        const fields = fieldsRef.current;
        const currentFieldValues = Object.values(fields);
        isSubmittedRef.current = true;
        isSubmittingRef.current = true;
        reRenderForm(Object.keys(errors).length ? errors : {});
        submitCountRef.current += 1;
        if (validationSchema) {
            fieldValues = getFieldsValues(fields);
            fieldErrors = await validateWithSchema(validationSchema, fieldValues);
            if (fieldErrors === undefined) {
                isSubmittingRef.current = false;
                reRenderForm(errors);
                return callback(combineFieldValues(fieldValues), e);
            }
        }
        else {
            const result = await currentFieldValues.reduce(async (previous, field) => {
                const resolvedPrevious = await previous;
                const { ref, ref: { name }, } = field;
                if (!fields[name])
                    return Promise.resolve(resolvedPrevious);
                const fieldError = await validateField(field, fields);
                const hasError = fieldError && fieldError[name];
                if (!hasError) {
                    resolvedPrevious.values[name] = getFieldValue(fields, ref);
                    return Promise.resolve(resolvedPrevious);
                }
                resolvedPrevious.errors = Object.assign({}, (resolvedPrevious.errors || {}), fieldError);
                return Promise.resolve(resolvedPrevious);
            }, Promise.resolve({
                errors: {},
                values: {},
            }));
            fieldErrors = result.errors;
            fieldValues = result.values;
        }
        isSubmittingRef.current = false;
        reRenderForm(fieldErrors || {});
        if (Object.values(fieldErrors).length) {
            errorsRef.current = fieldErrors;
            return;
        }
        callback(combineFieldValues(fieldValues), e);
    };
    const unSubscribe = () => {
        fieldsRef.current &&
            Object.values(fieldsRef.current).forEach((field) => {
                const { ref, options } = field;
                isRadioInput(ref.type) && Array.isArray(options)
                    ? options.forEach((fieldRef) => removeEventListener(fieldRef, true))
                    : removeEventListener(field, true);
            });
        fieldsRef.current = {};
        watchFieldsRef.current = {};
        errorsRef.current = {};
        isWatchAllRef.current = false;
        isSubmittedRef.current = false;
        isDirtyRef.current = false;
        touchedFieldsRef.current = [];
        reRenderForm({});
    };
    react.useEffect(() => unSubscribe, [mode]);
    return {
        register,
        handleSubmit,
        watch,
        unSubscribe,
        errors,
        formState: {
            dirty: isDirtyRef.current,
            isSubmitted: isSubmittedRef.current,
            submitCount: submitCountRef.current,
            touched: touchedFieldsRef.current,
            isSubmitting: isSubmittingRef.current,
        },
    };
}

module.exports = useForm;
